#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# License: GNU General Public License v2

# Author: thl-cmk[at]outlook[dot]com
# URL   : https://thl-cmk.hopto.org
# Date  : 2023-10-12
# File  : nvdct/lib/backends.py

# 2024-06-18: fixed host_exist returns always True if host was in host_cache, even with host=None

from _collections_abc import Mapping, Sequence
from abc import abstractmethod
from ast import literal_eval
from enum import Enum, unique
from pathlib import Path
from requests import session
from time import time_ns
from typing import Dict, List, Tuple
from sys import exit as sys_exit

from livestatus import MultiSiteConnection, SiteConfigurations, SiteId

from lib.utils import (
    CACHE_INTERFACES_DATA,
    ExitCodes,
    get_data_form_live_status,
    get_table_from_inventory,
    LOGGER,
    OMD_ROOT,
    PATH_INTERFACES,
)


def hosts_to_query(hosts: List[str]) -> Tuple[str, List[str]]:
    # WORKAROUND for: Apache HTTP Error 414: Request URI too long
    # https://httpd.apache.org/docs/current/mod/core.html#limitrequestfieldsize
    # Default:	LimitRequestFieldSize 8190
    # max seems to be 8013 (inventory)/7831 (interfaces), so 7800 is a save distance from it
    _max_str_len = 7800
    # 8190 - 8013 - 168 = 9 chrs  -> Inventory
    # 8190 - 7813 - 356 = 21 chrs -> Interfaces
    # host (19/19): http://localhost:80 -> sould not count
    # uri (57/60): /build/check_mk/api/1.0/domain-types/host/collections/all / /build/check_mk/api/1.0/domain-types/service/collections/all  # noqa: E501
    # query (77/231): ?query=%7B%22op%22%3A+%22~~%22%2C+%22left%22%3A+%22name%22%2C+%22right%22%3A+ / ?query=%7B%22op%22%3A+%22and%22%2C+%22expr%22%3A+%5B%7B%22op%22%3A+%22~%22%2C+%22left%22%3A+%22description%22%2C+%22right%22%3A+%22Interface+%22%7D%2C%7B%22op%22%3A+%22~~%22%2C+%22left%22%3A+%22host_name%22%2C+%22right%22%3A+%22%5E  # noqa: E501
    # collumns (34/65): &columns=name&columns=mk_inventory / &columns=host_name&columns=description&columns=long_plugin_output  # noqa: E501
    # %22 -> "
    # %24 -> $
    # %2C -> ,
    # %3A -> :
    # %5B -> [
    # %5D -> ]
    # %7B -> {
    # %7D -> }

    temp_hosts = []
    open_hosts = []
    for host in hosts:
        temp_hosts.append(f'^{host}$')
    hosts_str = '|'.join(temp_hosts)
    # 6 comes from 3 chrs ($|^) between hosts, coded as 9 chr(%24%7C%5E) in the URL
    # 3 are already in the str, so we need to add 6 for each host
    if len(hosts_str) > _max_str_len - len(hosts) * 6:
        open_hosts = hosts.copy()
        hosts_str = f'^{hosts[0]}$'
        hosts = hosts[1:]
        count = 1
        for host in hosts:
            count += 1
            if len(hosts_str) + 9 + len(host) + (6 * count) < _max_str_len:
                hosts_str = hosts_str + f'|^{host}$'
                open_hosts.remove(host)
            else:
                break

    LOGGER.debug(f'hosts len: {len(hosts_str) + (len(hosts) - len(open_hosts)) * 6}')
    LOGGER.debug(f'open hosts {open_hosts}')

    return hosts_str, open_hosts


@unique
class CacheItems(Enum):
    inventory = 'inventory'
    interfaces = 'interfaces'


class HostCache:
    def __init__(
            self,
            pre_fetch: bool,
            backend: str,
    ):
        self._cache: Dict = {}
        self._inventory_pre_fetch_list: List[str] = [
            PATH_INTERFACES,
        ]
        self._count: int = 0
        self._pre_fetch = pre_fetch
        self._backend = backend
        LOGGER.info('init HOST_CACHE')

    @abstractmethod
    def get_inventory_data(self, hosts: Sequence[str]) -> Dict[str, Dict | None]:
        """
        Args:
            hosts: the host name to return the inventory data for

        Returns:
            the inventory data as dictionary
        """
        raise NotImplementedError()

    @abstractmethod
    def get_interface_data(self, hosts: Sequence[str]) -> Dict[str, Dict | None]:
        """

        Args:
            hosts:  lit of host names to return the interface data for

        Returns:
            dictionary of the interface data with the item as key
        """
        raise NotImplementedError()

    @abstractmethod
    def host_exists(self, host: str) -> bool:
        raise NotImplementedError()

    @abstractmethod
    def get_hosts_by_label(self, label: str) -> List[str] | None:
        raise NotImplementedError()

    @abstractmethod
    def pre_fetch_hosts(self):
        raise NotImplementedError()

    @property
    def cache(self) -> Dict:
        return self._cache

    @property
    def backend(self) -> str:
        return self._backend

    @property
    def pre_fetch(self) -> bool:
        return self._pre_fetch

    def stop_host(self, host: str) -> None:
        if host not in self._cache:
            self._cache[host] = None

    def pre_fetch_cache(self, hosts: Sequence[str]) -> None:
        # pre fill inventory data
        self._count += 1
        _pre_query = time_ns()

        inventory_of_hosts: Mapping[str, Mapping | None] = self.get_inventory_data(hosts=hosts)
        LOGGER.debug(f'{(time_ns() - _pre_query) / 1e9}|{self._count:0>4}|inventory|{hosts}')

        if inventory_of_hosts:
            for host, inventory in inventory_of_hosts.items():
                if host not in self._cache:
                    self._cache[host] = {}
                self._cache[host][CacheItems.inventory.value] = {}
                self._cache[host][CacheItems.inventory.value].update({
                    entry: get_table_from_inventory(
                        inventory=inventory,
                        raw_path=entry
                    ) for entry in self._inventory_pre_fetch_list
                })

        _pre_query = time_ns()

        interfaces_of_hosts: Mapping[str, Mapping | None] = self.get_interface_data(hosts)

        for host, interfaces in interfaces_of_hosts.items():
            if host not in self._cache:
                self._cache[host] = {}
            if not self._cache[host].get(CacheItems.interfaces.value):
                self._cache[host][CacheItems.interfaces.value] = {}
            self._cache[host][CacheItems.interfaces.value][CACHE_INTERFACES_DATA] = interfaces

            LOGGER.debug(f'{(time_ns() - _pre_query) / 1e9}|{self._count:0>4}|items|{host}')

    def get_data(self, host: str, item: CacheItems, path: str) -> Dict[str, any] | None:
        if self.host_exists(host=host):
            if host not in self._cache:
                self._cache[host] = {}
                LOGGER.info(f'Host not in cache: {host}')
                self.pre_fetch_cache(hosts=[host])
            try:
                return self._cache[host][item.value][path]
            except (KeyError, TypeError):
                return None
        else:
            self._cache[host] = None
        return None

    def add_inventory_prefetch_path(self, path: str) -> None:
        self._inventory_pre_fetch_list = list(set(self._inventory_pre_fetch_list + [path]))


class HostCacheLiveStatus(HostCache):
    def __init__(self, pre_fetch: bool, backend: str = '[LIVESTATUS]'):
        super().__init__(pre_fetch, backend)
        if self.pre_fetch:
            self.pre_fetch_hosts()

    def get_raw_data(self, query: str) -> any:
        return get_data_form_live_status(query=query)

    def get_inventory_data(self, hosts: List[str]) -> Dict[str, Dict | None]:
        host_data: Dict[str, Dict | None] = {}
        # int host_data with None
        for host in hosts:
            host_data[host] = None
        open_hosts = hosts.copy()
        while open_hosts:
            hosts_str, open_hosts = hosts_to_query(open_hosts)
            query = (
                'GET hosts\n'
                'Columns: host_name mk_inventory\n'
                'OutputFormat: python3\n'
                f'Filter: host_name ~~ {hosts_str}\n'
            )
            data: Sequence[Tuple[str, bytes]] = self.get_raw_data(query=query)
            LOGGER.debug(f'{self.backend} data for hosts {hosts}: {data}')
            if data:
                for host, inventory in data:
                    if inventory == b'':
                        LOGGER.warning(f'{self.backend} Device: {hosts}: no inventory data found!')
                        continue
                    try:
                        host_data[host] = literal_eval(inventory.decode('utf-8'))
                    except SyntaxError as e:
                        LOGGER.exception(f'inventory: |{inventory!r}|')  # output raw data
                        LOGGER.exception(f'type: {type(inventory)}')
                        LOGGER.exception(f'exception: {e}')
                        continue
            else:
                LOGGER.warning(f'{self.backend} Device: {hosts}: no inventory data found!')
        return host_data

    def get_interface_data(self, hosts: List[str]) -> Dict[str, Dict | None]:
        # host_data: Dict[str, Dict[str, Dict[str, List[str]] ] | None] = {}
        host_data: Dict[str, any] = {}  # to make pylint happy
        # int host_data with None
        for host in hosts:
            host_data[host] = None
        open_hosts = hosts.copy()
        while open_hosts:
            hosts_str, open_hosts = hosts_to_query(open_hosts)
            query = (
                'GET services\n'
                'Columns: host_name description long_plugin_output\n'
                'Filter: description ~ ^Interface\n'
                f'Filter: host_name ~~ {hosts_str}\n'
                'OutputFormat: python3\n'
            )
            data: List[Tuple[str, str, str]] = self.get_raw_data(query=query)
            LOGGER.debug(f'{self.backend} data for host {hosts}: {data}')
            if data:
                for host, description, long_plugin_output in data:
                    if host_data.get(host) is None:
                        host_data[host] = {}
                    host_data[host][description[10:]] = {  # remove 'Interface ' from description
                        'long_plugin_output': long_plugin_output.split('\\n')
                    }
            else:
                LOGGER.warning(f'{self.backend} No Interfaces items found for hosts {hosts}')

        return host_data

    def host_exists(self, host: str) -> bool:
        query = (
            'GET hosts\n'
            'Columns: host_name\n'
            'OutputFormat: python3\n'
            f'Filter: host_name = {host}\n'
        )
        if self.cache.get(host) is not None:
            return True
        elif host in self.cache:
            return False
        # if self.pre_fetch:
        #     LOGGER.warning(f'{self.backend} pre_fetch host not found in cache {host}')
        #     return False
        data: Sequence[Sequence[str]] = self.get_raw_data(query=query)
        LOGGER.debug(f'{self.backend} data for host {host}: {data}')
        if [host] in data:
            LOGGER.debug(f'{self.backend} Host {host} found in CMK')
            return True

        LOGGER.warning(f'{self.backend} Host {host} not found in CMK')
        self.stop_host(host)

        return False

    def get_hosts_by_label(self, label: str) -> List[str] | None:
        query = (
            'GET hosts\n'
            'Columns: name\n'
            'OutputFormat: python3\n'
            f'Filter: label_names ~ {label}\n'
        )
        data: Sequence[Sequence[str]] = self.get_raw_data(query=query)
        LOGGER.debug(f'{self.backend} routing capable hosts: {data}')
        if data:
            hosts = []
            for host in data:
                hosts.append(host[0])
            return hosts

        LOGGER.debug(f'{self.backend} no routing capable hosts found')
        return None

    def pre_fetch_hosts(self):
        LOGGER.debug(f'{self.backend} pre_fetch_hosts')
        query = (
            'GET hosts\n'
            'Columns: host_name\n'
            'OutputFormat: python3\n'
        )
        data: Sequence[Sequence[str]] = self.get_raw_data(query=query)
        if data:
            for host in data:
                self._cache[host[0]] = {}
            LOGGER.debug(f'{self.backend} # of host found: {len(self.cache.keys())}')
        else:
            LOGGER.warning(f'{self.backend} no hosts found')


class HostCacheMultiSite(HostCacheLiveStatus):
    def __init__(self, pre_fetch: bool):
        self._backend = '[MULTISITE]'
        self.sites: SiteConfigurations = SiteConfigurations({})
        self.get_sites()
        LOGGER.info(f'{self.backend} Create livestatus connection(s)')
        self.c = MultiSiteConnection(self.sites)
        # self.c.set_prepend_site(False)  # is default
        # self.c.parallelize = True   # is default
        self.dead_sites = [site['site']['alias'] for site in self.c.dead_sites().values()]
        if self.dead_sites:
            dead_sites = ', '.join(self.dead_sites)
            LOGGER.warning(f'{self.backend} WARNING: use of dead site(s) {dead_sites} is disabled')
            self.c.set_only_sites(self.c.alive_sites())
        super().__init__(pre_fetch, self._backend)
        if self.pre_fetch:
            self.pre_fetch_hosts()

    def get_sites(self):
        # see: https://github.com/Checkmk/checkmk/blob/master/packages/cmk-livestatus-client/example_multisite.py

        sites_mk = Path(f'{OMD_ROOT}/etc/check_mk/multisite.d/sites.mk')
        socket_path = f'unix:{OMD_ROOT}/tmp/run/live'
        if sites_mk.exists():
            # make eval() "secure"
            # https://realpython.com/python-eval-function/#minimizing-the-security-issues-of-eval
            _code = compile(sites_mk.read_text(), "<string>", "eval")
            allowed_names = ['sites', 'update']
            for name in _code.co_names:
                if name not in allowed_names:
                    raise NameError(f'Use of {name} in {sites_mk.name} not allowed.')

            sites_raw: Dict = {}
            eval(sites_mk.read_text(), {"__builtins__": {}}, {"sites": sites_raw})

            for site, data in sites_raw.items():
                self.sites.update({site: {
                    'alias': data['alias'],
                    'timeout': data['timeout'],
                    # 'nagios_url': '/nagios/',
                }})
                if data['socket'] == ('local', None):
                    self.sites[site]['socket'] = socket_path
                else:
                    protocol, socket = data['socket']
                    address, port = socket['address']
                    self.sites[site]['socket'] = f'{protocol}:{address}:{port}'
                    self.sites[site]['tls'] = socket['tls']
                LOGGER.info(
                    f'{self.backend} Site {site} found, '
                    f'Socket: {self.sites[site]["socket"]}, '
                    f'TLS: {self.sites[site].get("tls", "N/A")}.'
                )
        else:
            self.sites.update({SiteId('local'): {
                'alias': 'Local site',
                'timeout': 5,
                # 'nagios_url': '/nagios/',
                'socket': socket_path
            }})

            LOGGER.critical(
                f'{self.backend} file {str(sites_mk.absolute())} not found. Fallback to'
                'local site only. Try -b RESTAPI if you have a distributed environment.'
            )

    def get_raw_data(self, query: str) -> object:
        return self.c.query(query=query)


class HostCacheRestApi(HostCache):
    def __init__(self, pre_fetch: bool, api_port: int):
        super().__init__(pre_fetch, '[RESTAPI]')
        LOGGER.debug(f'{self.backend} init backend')
        self._api_port = api_port
        try:
            self.__secret = Path(
                f'{OMD_ROOT}/var/check_mk/web/automation/automation.secret'
            ).read_text().strip('\n')
        except FileNotFoundError as e:
            LOGGER.exception(f'{self.backend} automation.secret not found, {e}')
            print(f'{self.backend} automation.secret not found, {e}')
            sys_exit(ExitCodes.AUTOMATION_SECRET_NOT_FOUND.value)
        self.__hostname = 'localhost'
        self.__site = OMD_ROOT.split('/')[-1]
        self.__api_url = f"http://{self.__hostname}:{self._api_port}/{self.__site}/check_mk/api/1.0"
        self.__user = 'automation'
        LOGGER.info(f'{self.backend} Create REST API session')
        self.__session = session()
        self.__session.headers['Authorization'] = f"Bearer {self.__user} {self.__secret}"
        self.__session.headers['Accept'] = 'application/json'

        if self.pre_fetch:
            self.pre_fetch_hosts()

    def get_inventory_data(self, hosts: List[str]) -> Dict[str, Dict | None]:
        LOGGER.debug(f'{self.backend} get_inventory_data {hosts}')
        host_data: Dict[str, Dict | None] = {}
        # init host_data with None
        for host in hosts:
            host_data[host] = None
        open_hosts = hosts.copy()
        while open_hosts:
            hosts_str, open_hosts = hosts_to_query(open_hosts)
            LOGGER.debug(f'{self.backend} open hosts: {open_hosts}, len: {len(open_hosts)}')
            query = '{"op": "~~", "left": "name", "right": "' + hosts_str + '"}'
            resp = self.__session.get(
                url=f"{self.__api_url}/domain-types/host/collections/all",
                params={
                    "query": query,
                    "columns": ['name', 'mk_inventory'],
                },
            )
            if resp.status_code == 200:
                LOGGER.debug(f'{self.backend} {resp.elapsed}|{self._count:0>4}|inventory|{hosts}')
                data = resp.json().get('value', [])
                for raw_host in data:
                    host = raw_host.get('extensions', {}).get('name')
                    if host:
                        host_data[host] = raw_host['extensions'].get('mk_inventory')
            else:
                LOGGER.info(
                    f'{self.backend} got no inventory data found!, status code {resp.status_code}'
                )
                LOGGER.debug(f'{self.backend} response query: {query}')
                LOGGER.debug(f'{self.backend} response text: {resp.text}')
                LOGGER.debug(f'{self.backend} response url: {resp.url}, len: {len(resp.url)}')

        return host_data

    def get_interface_data(self, hosts: List[str]) -> Dict[str, Dict | None]:
        LOGGER.debug(f'{self.backend} get_interface_data {hosts}')
        host_data: Dict[str, Dict | None] = {}
        # init host_data with None
        for host in hosts:
            host_data[host] = None
        open_hosts = hosts.copy()
        while open_hosts:
            hosts_str, open_hosts = hosts_to_query(open_hosts)

            query_host = f'{{"op": "~~", "left": "host_name", "right": "{hosts_str}"}}'
            query_item = '{"op": "~", "left": "description", "right": "Interface "}'
            query = f'{{"op": "and", "expr": [{query_item},{query_host}]}}'

            resp = self.__session.get(
                url=f"{self.__api_url}/domain-types/service/collections/all",
                params={
                    "query": query,
                    "columns": ['host_name', 'description', 'long_plugin_output'],
                },
            )

            if resp.status_code == 200:
                LOGGER.debug(f'{resp.elapsed}|{self._count:0>4}|items|{hosts}')

                data = resp.json().get('value', [])
                for raw_service in data:
                    LOGGER.debug(f'{self.backend} data for service : {raw_service}')
                    service = raw_service.get('extensions')
                    host, description, long_plugin_output = service.values()
                    host = service['host_name']
                    description = service['description']
                    if host_data.get(host) is None:
                        host_data[host] = {}
                    host_data[host][description[10:]] = {
                        'long_plugin_output': long_plugin_output.split('\\n')
                    }
            else:
                LOGGER.warning(
                    f'{self.backend} got no interface data, response code {resp.status_code}'
                )
                LOGGER.debug(f'{self.backend} response query: {query}')
                LOGGER.debug(f'{self.backend} response text: {resp.text}')
                LOGGER.debug(f'{self.backend} response url: {resp.url}, len: {len(resp.url)}')

        return host_data

    def host_exists(self, host: str) -> bool:
        LOGGER.debug(f'{self.backend} host_exists {host}')
        if self.cache.get(host) is not None:
            LOGGER.debug(f'{self.backend} host found in cache {host}')
            return True
        elif host in self.cache:
            return False

        # if self.pre_fetch:
        #     LOGGER.warning(f'{self.backend} pre_fetch host not found in cache {host}')
        #     return False
        query = '{"op": "=", "left": "name", "right": "' + host + '"}'
        resp = self.__session.get(
                url=f"{self.__api_url}/domain-types/host/collections/all",
                params={
                    "query": query,
                    "columns": ['name'],
                },
            )
        if resp.status_code == 200:
            LOGGER.debug(f'{resp.elapsed}|{self._count:0>4}|name|{host}')
            try:
                data = resp.json()['value'][0]['extensions']['name']
                LOGGER.debug(f'{self.backend} data for host {host}: {resp.json()}')
            except IndexError:
                LOGGER.warning(f'Host {host} not found in CMK')
                self.stop_host(host)
                return False
            if data == host:
                return True
        else:
            LOGGER.warning(f'{self.backend} response: {resp.status_code}')
        return False

    def get_hosts_by_label(self, label: str) -> List[str] | None:
        LOGGER.debug(f'{self.backend} get_hosts_by_label {label}')
        query = '{"op": "~", "left": "label_names", "right": "' + label + '"}'
        resp = self.__session.get(
            url=f"{self.__api_url}/domain-types/host/collections/all",
            params={
                "query": query,
                "columns": ['name'],
            },
        )
        if resp.status_code == 200:
            LOGGER.debug(f'{self.backend} data for routing_capable: {resp.json()}')
            LOGGER.debug(f'{resp.elapsed}|{self._count:0>4}|name|routing_capable')
            try:
                data = resp.json().get('value')
                hosts = []
                for host in data:
                    hosts.append(host['extensions']['name'])
                LOGGER.debug(f'{self.backend} host list {hosts}')
                return hosts
            except IndexError:
                LOGGER.debug(f'{self.backend} no routing capable hosts found')
                return None
        else:
            LOGGER.warning(f'{self.backend} response: {resp.status_code}')
        return None

    def pre_fetch_hosts(self):
        LOGGER.debug(f'{self.backend} pre_fetch_hosts')

        resp = self.__session.get(
            url=f"{self.__api_url}/domain-types/host/collections/all",
            params={
                "columns": ['name'],
            },
        )
        if resp.status_code == 200:
            data = resp.json().get('value', [])
            for raw_host in data:
                host = raw_host.get('extensions', {}).get('name')
                if host:
                    self._cache[host] = {}
            LOGGER.debug(f'{self.backend} # of host found: {len(self.cache.keys())}')
        else:
            LOGGER.warning(f'{self.backend} respons: {resp.text}')


class HostCacheFileSystem(HostCache):
    def __init__(self, pre_fetch: bool):
        super().__init__(pre_fetch, '[FILESYSTEM]')

    def get_inventory_data(self, hosts: Sequence[str]) -> Dict[str, Dict | None]:
        host_data: Dict[str, Dict | None] = {}
        __inventory_path = 'var/check_mk/inventory'
        for host in hosts:
            # init host_data with None
            host_data[host] = None
            inventory_file: Path = Path(f'{OMD_ROOT}/{__inventory_path}/{host}')
            try:
                data = literal_eval(inventory_file.read_text())
            except FileNotFoundError:
                LOGGER.warning(
                    msg=f'{self.backend} Device: {host}: not found in inventory data path!'
                )
                continue

            LOGGER.debug(f'{self.backend} data for host {host}: {data}')
            host_data[host] = data

        return host_data

    def get_interface_data(self, hosts: Sequence[str]) -> Dict[str, Dict | None]:
        """
        Sample autochecks data, we keep only the item
        [
         {'check_plugin_name': 'if64', 'item': 'Fa0', 'parameters': {'discovered_oper_status': ['2'], ...}},\n  # noqa: E501
         {'check_plugin_name': 'if64', 'item': 'Fa1/0/1', 'parameters': {'discovered_oper_status': ['1'], ...}},\n  # noqa: E501
         {'check_plugin_name': 'if64', 'item': 'Fa1/0/10', 'parameters': {'discovered_oper_status': ['2'], ...}},\n  # noqa: E501
         {'check_plugin_name': 'if64', 'item': 'Fa1/0/11', 'parameters': {'discovered_oper_status': ['2'], ...}},\n  # noqa: E501
         {'check_plugin_name': 'if64', 'item': 'Fa1/0/12', 'parameters': {'discovered_oper_status': ['1'], ...}}\n  # noqa: E501
        ]

        Args:
            hosts: list of names of the host objects in cmk to fetch the data for

        Returns:
            List of interface service items

        """
        host_data: Dict[str, Dict | None] = {}
        __autochecks_path = 'var/check_mk/autochecks'
        for host in hosts:
            # init host_data with None
            host_data[host] = None
            autochecks_file: Path = Path(f'{OMD_ROOT}/{__autochecks_path}/{host}.mk')
            items: Dict[str, object] = {}
            try:
                data: Sequence[Mapping[str, str]] = literal_eval(autochecks_file.read_text())
            except FileNotFoundError:
                LOGGER.warning(
                    msg=f'{self.backend} Device: {host}: not found in auto checks path!'
                )
                continue
            LOGGER.debug(f'{self.backend} data for host {host}: {data}')

            for service in data:
                if service['check_plugin_name'] in ['if64']:
                    items[service['item']] = {}

            if items:
                LOGGER.info(
                    msg=f'{self.backend} Interfaces items found: {len(items)} an host {host}'
                )
                host_data[host] = items
            else:
                LOGGER.warning(
                    msg=f'{self.backend} No Interfaces items found for host {host}'
                )
        return host_data

    def host_exists(self, host: str) -> bool:
        # always True, don't works in distributed environments as autocheck files
        # are only locally available
        return False

    def get_hosts_by_label(self, label: str) -> List[str] | None:
        # not implemented, no (good) way to get a list of hosts from file system
        return None

    def pre_fetch_hosts(self):
        pass
