#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# License: GNU General Public License v2

# Author: thl-cmk[at]outlook[dot]com
# URL   : https://thl-cmk.hopto.org
# Date  : 2023-10-12
# File  : nvdct/lib/settings.py

# fixed path to default user data file

from _collections_abc import Mapping
from ipaddress import AddressValueError, IPv4Address, IPv4Network, NetmaskValueError
from logging import CRITICAL, FATAL, ERROR, WARNING, INFO, DEBUG
from os import environ
from pathlib import Path
from sys import exit as sys_exit
from time import strftime
from typing import Dict, List, NamedTuple

from lib.utils import (
    ExitCodes,
    get_data_from_toml,
    get_local_cmk_api_port,
    get_local_cmk_version,
    LOGGER,
    NVDCT_VERSION,
    SCRIPT,
    TIME_FORMAT,
    USER_DATA_FILE,
    Layer
)


class Emblems(NamedTuple):
    host_node: str
    ip_address: str
    ip_network: str
    l3v4_replace: str
    l3v4_summarize: str
    service_node: str


class Thickness(NamedTuple):
    speed: int
    thickness: int


class StaticConnection(NamedTuple):
    host: str
    local_port: str
    neighbour_port: str
    neighbour: str
    label: str


class Wildcard(NamedTuple):
    int_ip_address: int
    int_wildcard: int
    ip_address: str
    wildcard: str
    bit_pattern: int


class Settings:
    def __init__(
            self,
            cli_args: Mapping[str, object],
    ):
        self.__omd_root = environ['OMD_ROOT']
        self.__path_to_if_table = 'networking,interfaces'
        self.__topology_file_name = 'network_data.json'
        self.__topology_save_path = 'var/topology_data'
        self.__topology_save_path_cmk_2_3 = 'var/check_mk/topology/data'
        # cli defaults
        self.__settings = {
            # 'api_port': 80,
            'backend': 'MULTISITE',
            'check_user_data_only': False,
            'default': False,
            'dont_compare': False,
            'keep': False,
            'keep_domain': False,
            'layers': ['CDP'],
            'log_file': f'{self.omd_root}/var/log/nvdct.log',
            'log_level': 'WARNING',
            'log_to_stdout': False,
            'lowercase': False,
            'min_age': 0,
            # 'new_format': False,
            'output_directory': None,
            'prefix': None,
            'quiet': False,
            'pre_fetch': False,
            'seed_devices': [],
            'skip_l3_if': False,
            'skip_l3_ip': False,
            'time_format': TIME_FORMAT,
            'uppercase': False,
            'user_data_file': f'{self.omd_root}/local/bin/nvdct/conf/{USER_DATA_FILE}',
            'version': False,
        }
        # args in the form {'s, __seed_devices': 'CORE01', 'p, __path_in_inventory': None, ... }}
        # we will remove 's, __'
        self.__args = (
            {k.split(',')[-1].strip(' ').strip('_'): v for k, v in cli_args.items() if v}
        )

        if self.__args.get('version'):
            print(f'{Path(SCRIPT).name} version: {NVDCT_VERSION}')
            sys_exit(ExitCodes.OK.value)

        self.__user_data = get_data_from_toml(
            file=self.__args.get('user_data_file', self.user_data_file)
        )

        if self.__args.get('check_user_data_only'):
            LOGGER.info(msg=f'Could read/parse the user data from {self.user_data_file}')
            print(f'Could read/parse the user data from {self.user_data_file}')
            sys_exit(ExitCodes.OK.value)

        # defaults -> overridden by toml -> overridden by cli
        self.__settings.update(self.__user_data.get('SETTINGS', {}))
        self.__settings.update(self.__args)

        if self.layers:
            layers = list(set(self.layers))
            if len(layers) != len(self.layers):
                LOGGER.error(
                    msg='-l/--layers options must be unique. Don\'t use any layer more than once.'
                )
                print('-l/--layers options must be unique. Don\'t use any layer more than once.')
                sys_exit(ExitCodes.BAD_OPTION_LIST.value)

        self.__api_port: int | None = None
        self.__new_format: bool | None = None

        # init user data with defaults
        self.__custom_layers: List[StaticConnection] | None = None
        self.__drop_host: List[str] | None = None
        self.__drop_host_regex: List[str] | None = None
        self.__host_map: Dict[str, str] | None = None
        self.__l3v4_ignore_hosts: List[str] | None = None
        self.__l3v4_ignore_ip: List[IPv4Network] | None = None
        self.__l3v4_ignore_wildcard: List[Wildcard] | None = None
        self.__l3v4_replace: Dict[str, str] | None = None
        self.__l3v4_summarize: List[IPv4Network] | None = None
        self.__map_speed_to_thickness: List[Thickness] | None = None
        self.__protected_topologies: List[str] | None = None
        self.__seed_devices: List[str] | None = None
        self.__static_connections: List[StaticConnection] | None = None
        self.__emblems: Emblems | None = None

    #
    # CLI settings
    #
    @property  # --api-port
    def api_port(self) -> int:
        if self.__api_port is None:
            if self.__settings.get('api_port'):
                self.__api_port = int(self.__settings.get('api_port'))
            else:
                self.__api_port = get_local_cmk_api_port()
            if self.__api_port is None:
                self.__api_port = 80

        return self.__api_port

    @property  # -b --backend
    def backend(self) -> str:
        return str(self.__settings['backend'])

    @property  # --check-user-data-only
    def check_user_data_only(self) -> bool:
        return bool(self.__settings['check_user_data_only'])

    @property  # -d --default
    def default(self) -> bool:
        return bool(self.__settings['default'])

    @property  # --dont-compare
    def dont_compare(self) -> bool:
        return bool(self.__settings['dont_compare'])

    @property  # --keep
    def keep(self) -> int | None:
        if isinstance(self.__settings['keep'], int):
            return max(self.__settings['keep'], 0)
        return None

    @property  # --keep-domain
    def keep_domain(self) -> bool:
        return bool(self.__settings['keep_domain'])

    @property  # --layers
    def layers(self) -> List[str]:
        return self.__settings['layers']

    @property  # --log-file
    def log_file(self) -> str:
        return str(self.__settings['log_file'])

    @property  # --log-level
    def loglevel(self) -> int:
        match self.__settings['log_level']:
            case 'DEBUG':
                return DEBUG
            case 'INFO':
                return INFO
            case 'WARNING':
                return WARNING
            case 'ERROR':
                return ERROR
            case 'FATAL':
                return FATAL
            case 'CRITICAL':
                return CRITICAL
            case 'OFF':
                return -1
            case _:
                return WARNING

    @property  # --log-to-stdout
    def log_to_stdtout(self) -> bool:
        return bool(self.__settings['log_to_stdout'])

    @property  # --lowercase
    def lowercase(self) -> bool:
        return bool(self.__settings['lowercase'])

    @property  # --min-age
    def min_age(self) -> int:
        if isinstance(self.__settings['min_age'], int):
            return max(self.__settings['min_age'], 0)
        else:
            return 0

    @property  # --new-format
    def new_format(self) -> bool:
        if self.__new_format is None:
            if self.__settings.get('new_format') is True:
                self.__new_format = True
            elif get_local_cmk_version().startswith('2.2'):
                self.__new_format = False
            else:
                self.__new_format = True

        return self.__new_format

    @property  # --output-directory
    def output_directory(self) -> str:
        # init output directory with current time if not set
        if not self.__settings['output_directory']:
            self.__settings['output_directory'] = f'{strftime(self.__settings["time_format"])}'
        return str(self.__settings['output_directory'])

    @property  # --prefix
    def prefix(self) -> str | None:
        if self.__settings['prefix'] is not None:
            return str(self.__settings['prefix'])
        return None

    @property  # --pre-fill-cache
    def pre_fetch(self) -> bool:
        return bool(self.__settings['pre_fetch'])

    @property  # --quiet
    def quiet(self) -> bool:
        return bool(self.__settings['quiet'])

    @property  # --seed-devices
    def seed_devices(self) -> List[str]:
        if self.__seed_devices is None:
            self.__seed_devices = list(set(str(host) for host in (
                self.__user_data.get('SEED_DEVICES', []) + self.__settings.get('seed_devices', [])
            )))
        return self.__seed_devices

    @property  # --skip-l3-if
    def skip_l3_if(self) -> bool:
        return bool(self.__settings['skip_l3_if'])

    @property  # --skip-l3-ip
    def skip_l3_ip(self) -> bool:
        return bool(self.__settings['skip_l3_ip'])

    @property  # --time-format
    def time_format(self) -> str:
        return str(self.__settings['time_format'])

    @property  # --uppercase
    def uppercase(self) -> bool:
        return bool(self.__settings['uppercase'])

    @property  # --user-data-file
    def user_data_file(self) -> str:
        return str(self.__settings['user_data_file'])

    @property  # --version
    def version(self) -> bool:
        return bool(self.__settings['version'])

    #
    #  user data setting
    #
    @property
    def custom_layers(self) -> List[Layer]:
        if self.__custom_layers is None:
            self.__custom_layers = []
            for _layer in self.__user_data.get('CUSTOM_LAYERS', []):
                try:
                    self.__custom_layers.append(Layer(
                        path=_layer['path'],
                        columns=_layer['columns'],
                        label=_layer['label'],
                        host_label=_layer['host_label']
                    ))
                except KeyError:
                    LOGGER.error(
                        f'Invalid entry in CUSTOM_LAYERS -> {_layer} -> ignored'
                    )
                    continue
            LOGGER.critical(
                f'Valid entries in CUSTOM_LAYERS found: {len(self.__custom_layers)}/'
                f'{len(self.__user_data.get("CUSTOM_LAYERS", []))}'
            )
        return self.__custom_layers

    @property
    def drop_hosts(self) -> List[str]:
        if self.__drop_host is None:
            self.__drop_host = [str(host) for host in self.__user_data.get('DROP_HOSTS', [])]
        return self.__drop_host

    @property
    def drop_hosts_regex(self) -> List[str]:
        if self.__drop_host_regex is None:
            self.__drop_host_regex = [str(host) for host in self.__user_data.get(
                'DROP_HOSTS_REGEX', []
            )]
        return self.__drop_host_regex

    @property
    def emblems(self) -> Emblems:
        if self.__emblems is None:
            raw_emblems = self.__user_data.get('EMBLEMS', {})
            self.__emblems = Emblems(
                host_node=str(raw_emblems.get('host_node', 'icon_missing')),
                ip_address=str(raw_emblems.get('ip_address', 'ip-address_80')),
                ip_network=str(raw_emblems.get('ip_network', 'ip-network_80')),
                l3v4_replace=str(raw_emblems.get('l3v4_replace', 'icon_plugins_cloud')),
                l3v4_summarize=str(raw_emblems.get('l3v4_summarize', 'icon_aggr')),
                service_node=str(raw_emblems.get('service_node', 'icon_missing')),
            )
        return self.__emblems

    @property
    def host_map(self) -> Dict[str, str]:
        if self.__host_map is None:
            self.__host_map = {
                str(host): str(replace_host) for host, replace_host in self.__user_data.get(
                    'HOST_MAP', {}
                ).items()
            }
        return self.__host_map

    @property
    def l3v4_ignore_hosts(self) -> List[str]:
        if self.__l3v4_ignore_hosts is None:
            self.__l3v4_ignore_hosts = [str(host) for host in self.__user_data.get(
                'L3V4_IGNORE_HOSTS', []
            )]
        return self.__l3v4_ignore_hosts

    @property
    def l3v4_ignore_ips(self) -> List[IPv4Network]:
        if self.__l3v4_ignore_ip is None:
            self.__l3v4_ignore_ip = []
            for ip_network in self.__user_data.get('L3V4_IGNORE_IP', []):
                try:
                    self.__l3v4_ignore_ip.append(IPv4Network(ip_network, strict=False))
                except (AddressValueError, NetmaskValueError):
                    LOGGER.error(
                        f'Invalid entry in L3V4_IGNORE_IP found: {ip_network} -> ignored'
                    )
                    continue
            LOGGER.info(
                f'Valid entries in L3V4_IGNORE_IP found: {len(self.__l3v4_ignore_ip)}/'
                f'{len(self.__user_data.get("L3V4_IGNORE_IP", []))}'
            )

        return self.__l3v4_ignore_ip

    @property
    def l3v4_ignore_wildcard(self) -> List[Wildcard]:
        if self.__l3v4_ignore_wildcard is None:
            self.__l3v4_ignore_wildcard = []
            for entry in self.__user_data.get('L3V4_IRNORE_WILDCARD', []):
                try:
                    ip_address, wildcard = entry
                except ValueError:
                    LOGGER.error(
                        f'Invalid entry in L3V4_IRNORE_WILDCARD -> {entry} -> ignored'
                    )
                    continue
                try:
                    # 0.0.127.0 -> 255.255.128.255
                    inverted_wildcard = '.'.join(
                        [str(255 - int(octet)) for octet in wildcard.split('.')]
                    )
                    self.__l3v4_ignore_wildcard.append(Wildcard(
                        int_ip_address=int(IPv4Address(ip_address)),
                        int_wildcard=int(IPv4Address(inverted_wildcard)),
                        ip_address=ip_address,
                        wildcard=wildcard,
                        bit_pattern=int(IPv4Address(ip_address)) & int(
                            IPv4Address(inverted_wildcard)
                        )
                    ))
                except (AddressValueError, NetmaskValueError):
                    LOGGER.error(
                        f'Invalid entry in L3V4_IRNORE_WILDCARD -> {entry} -> ignored'
                    )
                    continue
            LOGGER.info(
                f'Valid entries in L3V4_IRNORE_WILDCARD found: {len(self.__l3v4_ignore_wildcard)}/'
                f'{len(self.__user_data.get("L3V4_IRNORE_WILDCARD", []))}'
            )
        return self.__l3v4_ignore_wildcard

    @property
    def l3v4_replace(self) -> Dict[str, str]:
        if self.__l3v4_replace is None:
            self.__l3v4_replace = {}
            for ip_network, node in self.__user_data.get('L3V4_REPLACE', {}).items():
                try:
                    _ip_network = IPv4Network(ip_network)  # noqa: F841
                except (AddressValueError, NetmaskValueError):
                    LOGGER.error(
                        f'Invalid entry in L3V4_REPLACE found: {ip_network} -> ignored'
                    )
                    continue
                self.__l3v4_replace[ip_network] = str(node)
            LOGGER.info(
                f'Valid entries in L3V4_REPLACE found: {len(self.__l3v4_replace)}/'
                f'{len(self.__user_data.get("L3V4_REPLACE", {}))}'
            )
        return self.__l3v4_replace

    @property
    def l3v4_summarize(self) -> List[IPv4Network]:
        if self.__l3v4_summarize is None:
            self.__l3v4_summarize = []
            for ip_network in self.__user_data.get('L3V4_SUMMARIZE', []):
                try:
                    self.__l3v4_summarize.append(IPv4Network(ip_network, strict=False))
                except (AddressValueError, NetmaskValueError):
                    LOGGER.error(
                        f'Invalid entry in L3V4_SUMMARIZE -> {ip_network} -> ignored'
                    )
                    continue
            LOGGER.info(
                f'Valid entries in L3V4_SUMMARIZE found: {len(self.__l3v4_summarize)}/'
                f'{len(self.__user_data.get("L3V4_SUMMARIZE", []))}'
            )
        return self.__l3v4_summarize

    @property
    def map_speed_to_thickness(self) -> List[Thickness]:
        if self.__map_speed_to_thickness is None:
            self.__map_speed_to_thickness = []
            map_speed_to_thickness = self.__user_data.get(
                'MAP_SPEED_TO_THICKNESS', {}
            )
            for speed, thickness in map_speed_to_thickness.items():
                try:
                    self.__map_speed_to_thickness.append(Thickness(
                        speed=int(float(speed)),
                        thickness=int(thickness),
                    ))
                except ValueError:
                    LOGGER.error(
                        f'Invalid entry in MAP_SPEED_TO_THICKNESS -> {speed}={thickness} -> ignored'
                    )
                    continue
            LOGGER.info(
                f'Valid entries in MAP_SPEED_TO_THICKNESS found: {len(self.__map_speed_to_thickness)}'  # noqa: E501
                f'/{len(self.__user_data.get("MAP_SPEED_TO_THICKNESS", []))}'
            )
        return self.__map_speed_to_thickness

    @property
    def protected_topologies(self) -> List[str]:
        if self.__protected_topologies is None:
            self.__protected_topologies = [str(topology) for topology in self.__user_data.get(
                'PROTECTED_TOPOLOGIES', []
            )]
        return self.__protected_topologies

    @property
    def static_connections(self) -> List[StaticConnection]:
        if self.__static_connections is None:
            self.__static_connections = []
            for connection in self.__user_data.get('STATIC_CONNECTIONS', []):
                try:
                    host, local_port, neighbour_port, neighbour, label = connection
                except ValueError:
                    LOGGER.error(
                        f'Worng entry in STATIC_CONNECTIONS -> {connection} -> ignored'
                    )
                    continue
                self.__static_connections.append(StaticConnection(
                    host=str(host),
                    local_port=str(local_port),
                    neighbour_port=str(neighbour_port),
                    neighbour=str(neighbour),
                    label=str(label),
                ))
            LOGGER.info(
                f'Valid entries in STATIC_CONNECTIONS found: {len(self.__static_connections)}/'
                f'{len(self.__user_data.get("STATIC_CONNECTIONS", []))}'
            )
        return self.__static_connections

    #
    # all other settings
    #
    @property
    def omd_root(self) -> str:
        return self.__omd_root

    @property
    def topology_save_path(self) -> str:
        if self.new_format:
            return self.__topology_save_path_cmk_2_3

        return self.__topology_save_path

    @property
    def topology_file_name(self) -> str:
        return self.__topology_file_name
