#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# License: GNU General Public License v2

# Author: thl-cmk[at]outlook[dot]com
# URL   : https://thl-cmk.hopto.org
# Date  : 2024-06-09
# File  : lib/topologies.py

from _collections_abc import Mapping, Sequence
from ipaddress import IPv4Address, IPv4Network
from typing import Dict, List

from lib.backends import CACHE_INTERFACES_DATA, CacheItems, HostCache, PATH_INTERFACES
from lib.settings import Thickness, Wildcard
from lib.utils import LOGGER


class NvObjects:
    def __init__(self) -> None:
        self.nv_objects: Dict[str, any] = {}

    def add_host_object(
        self,
        host: str,
        host_cache: HostCache,
        emblem: str | None = None
    ) -> None:
        if host not in self.nv_objects:
            link: Dict = {}
            metadata: Dict = {}
            # LOGGER.debug(f'host: {host}, {host_cache.host_exists(host=host)}')
            if host_cache.host_exists(host=host) is True:
                LOGGER.debug(f'host: {host} exists')
                link = {'core': host}
            else:
                if emblem is not None:
                    metadata = {
                        'images': {
                            'emblem': emblem,  # node image
                            # 'icon': 'icon_tick',  # status image, top left from emblem
                        },
                        **({'tooltip': {
                            'quickinfo': [{'name': 'Host node', 'value': host}]
                        }} if not link else {})
                    }

            self.nv_objects[host] = {
                'name': host,
                'link': link,
                'metadata': metadata,
            }
            LOGGER.debug(f'host: {host}, link: {link}, metadata: {metadata}')

    def add_service_object(
        self,
        host: str,
        service: str,
        host_cache: HostCache,
        emblem: str | None = None,
        metadata: Dict | None = None,
        name: str | None = None,
        item: str | None = None,
    ) -> None:
        if metadata is None:
            metadata = {}
        if name is None:
            name = service
        speed = None

        self.add_host_object(host=host, host_cache=host_cache)
        service_object = f'{service}@{host}'
        if service_object not in self.nv_objects:
            link: Dict = {}
            if item is None:
                item = get_service_by_interface(host, service.lstrip('0'), host_cache)
            if item and host_cache.host_exists(host=host):
                service_long = f'Interface {item}'
                link = {'core': [host, service_long]}
                if op_data := get_operational_interface_data(host, item, host_cache):
                    metadata.update(op_data)
                    speed = op_data.get('op_speed_int')
            elif emblem is not None:
                metadata.update({
                    'images': {
                        'emblem': emblem,  # node image
                        # 'icon': 'icon_tick',  # status image, top left from emblem
                    },
                    **({'tooltip': {'quickinfo': [
                        {'name': 'Service node', 'value': item}
                    ]}} if not link else {})
                })
            if isinstance(speed, int):
                metadata.update({'native_speed': speed})

            self.nv_objects[service_object] = {
                'name': name,
                'link': link,
                'metadata': metadata,
            }
        elif metadata is not {}:
            self.nv_objects[service_object]['metadata'].update(metadata)

    def add_ipv4_address_object(
        self,
        host: str,
        ipv4_address: str,
        emblem: str,
        interface: str | None,
    ) -> None:
        if interface is not None:
            service_object = f'{ipv4_address}@{interface}@{host}'
        else:
            service_object = f'{ipv4_address}@{host}'

        if service_object not in self.nv_objects:
            self.nv_objects[service_object] = {
                'name': ipv4_address,
                'link': {},
                'metadata': {
                    'images': {
                        'emblem': emblem,
                        # 'icon': 'icon_tick',
                    },
                    'tooltip': {
                        'quickinfo': [
                            {'name': 'IP-Address', 'value': ipv4_address},
                        ]
                    }
                }
            }

    def add_ipv4_network_object(self, network: str, emblem: str, ) -> None:
        if network not in self.nv_objects:
            self.nv_objects[network] = {
                'name': network,
                'link': {},
                'metadata': {
                    'images': {
                        'emblem': emblem,  # node image
                        # 'icon': 'icon_tick',  # status image, top left from emblem
                    },
                    'tooltip': {
                        'quickinfo': [
                            {'name': 'IP-Network', 'value': network},
                        ]
                    }
                }
            }

    def add_tooltip_quickinfo(self, nv_object: str, name: str, value: str) -> Dict:
        metadata = self.nv_objects[nv_object]['metadata']
        if metadata.get('tooltip') is None:
            metadata['tooltip'] = {}
        if metadata['tooltip'].get('quickinfo') is None:
            metadata['tooltip']['quickinfo'] = []

        metadata['tooltip']['quickinfo'].append({
            'name': name,
            'value': value,
        })

        return metadata

    def add_icon_to_object(self, nv_object: str, icon: str) -> None:
        if nv_object in self.nv_objects:
            if not self.nv_objects[nv_object]['metadata'].get('images'):
                self.nv_objects[nv_object]['metadata']['images'] = {}
            self.nv_objects[nv_object]['metadata']['images']['icon'] = icon


class NvConnections:
    def __init__(self) -> None:
        self.nv_connections: List = []

    def add_connection(self, left: str, right: str):
        connection = [left, right]
        connection.sort()
        if [connection] not in self.nv_connections:
            self.nv_connections.append([connection])

    def add_meta_data_to_connections(
        self,
        nv_objects: NvObjects,
        speed_map: Sequence[Thickness],
    ):
        for connection in self.nv_connections:
            warning = False
            left, right = connection[0]
            metadata: Dict[str, Dict] = {
                'line_config': {
                    'css_styles': {
                        'stroke-dasharray': 'unset'
                    }
                }
            }
            left_speed = nv_objects.nv_objects[left].get('metadata', {}).get('op_speed_int')
            right_speed = nv_objects.nv_objects[right].get('metadata', {}).get('op_speed_int')

            left_speed_str = nv_objects.nv_objects[left].get('metadata', {}).get('op_sped_str')
            right_speed_str = nv_objects.nv_objects[right].get('metadata', {}).get('op_sped_str')

            left_duplex = nv_objects.nv_objects[left].get('metadata', {}).get('duplex')
            right_duplex = nv_objects.nv_objects[right].get('metadata', {}).get('duplex')

            left_native_vlan = nv_objects.nv_objects[left].get('metadata', {}).get('native_vlan')
            right_native_vlan = nv_objects.nv_objects[right].get('metadata', {}).get('native_vlan')

            if right_speed and left_speed:
                right_thickness = map_speed_to_thickness(right_speed, speed_map)
                # left_thickness = map_speed_to_thickness(left_speed, speed_map)
                metadata['line_config']['thickness'] = right_thickness

                if right_speed != left_speed:
                    warning = True
                    metadata = add_tooltip_html(
                        metadata, 'Speed', left, left_speed_str, right, right_speed_str
                    )
                    # metadata = add_tooltip_quickinfo(metadata, 'Warning', 'Speed mismatch')
                    # metadata = add_tooltip_quickinfo(metadata, left, left_speed_str)
                    # metadata = add_tooltip_quickinfo(metadata, right, right_speed_str)

                    LOGGER.warning(
                        f'Connection with speed mismatch: {left} (speed: {left_speed_str})'
                        f'<->{right} (speed: {right_speed_str})'
                    )

            # for duplex/native vlan it might be a good idea to change left/right
            # value as they are reported by CDP as neighbour states
            if left_duplex and right_duplex:
                if left_duplex != right_duplex:
                    warning = True

                    metadata = add_tooltip_html(
                        metadata, 'Duplex', left, left_duplex, right, right_duplex
                    )

                    LOGGER.warning(
                        f'Connection with duplex mismatch: {left} (duplex: {left_duplex})'
                        f'<->{right} (duplex: {right_duplex})'
                    )
            if left_native_vlan and right_native_vlan:
                if left_native_vlan != right_native_vlan:
                    warning = True

                    metadata = add_tooltip_html(
                        metadata, 'Native<br>VLAN', left, left_native_vlan, right, right_native_vlan
                    )

                    LOGGER.warning(
                        f'Connection with native vlan mismatch: '
                        f'{left} (vlan: {left_native_vlan})<->{right} (vlan: {right_native_vlan})'
                    )
            if warning:
                metadata['line_config'].update({
                        'color': 'red',
                        'thickness': 5,
                })
                metadata['line_config']['css_styles']['stroke-dasharray'] = '10 5'
                nv_objects.add_icon_to_object(left, 'icon_warning')
                nv_objects.add_icon_to_object(right, 'icon_warning')
                close_tooltip_html(metadata)
            connection.append(metadata)


def map_speed_to_thickness(speed_to_map: int, speed_map: Sequence[Thickness]) -> int:
    thickness: int = 1  # use in case of empty MAP_SPEED_TO_THICKNESS
    for speed, thickness in speed_map:
        if speed_to_map <= speed:
            break
    return thickness


def get_operational_interface_data(
    host: str,
    item: str,
    host_cache: HostCache,

) -> Dict[str, str | int] | None:
    unit_to_bits_per_second = {
        'Bit/s': 1,
        'kBit/s': 1000,
        'MBit/s': 1000000,
        'GBit/s': 1000000000,
    }

    # get dict of interfaces with the item as key
    interface_data: Dict[str, any] | None = host_cache.get_data(
        host=host, item=CacheItems.interfaces, path=CACHE_INTERFACES_DATA
    )
    try:
        raw_opdata = interface_data[item]['long_plugin_output']
    except (KeyError, TypeError):
        return None

    if raw_opdata:
        opdata: Dict[str, str | int] = {}
        for _entry in raw_opdata:
            try:
                key, value = _entry.split(': ', 1)  # split only at the first colon
            except ValueError:
                continue
            value = value.strip(' ')
            match key:
                case 'MAC':
                    if len(value) == 17:  # valid MAC: 6C:DD:30:DD:51:8B'
                        opdata['mac'] = value
                case 'Speed':
                    try:           # *_ -> ignore rest of string, i.e: (expected: 1 GBit/s)WARN
                        speed, unit, *_ = value.split(' ')
                    except ValueError:
                        pass
                    else:
                        opdata['op_sped_str'] = f'{speed} {unit}'
                        opdata['op_speed_int'] = int(float(speed) * unit_to_bits_per_second[unit])

        return opdata
    return None


def get_service_by_interface(host: str, interface: str, host_cache: HostCache) -> str | None:
    """
    Returns:
        Tuple of interface item
    """

    _alternate_if_name = {
        'ethernet': 'eth',
        # 'fastethernet': 'Fa',
        # 'gigabitethernet': 'gi',
        # 'tengigabitethernet': 'te',
        # 'fortygigabitethernet': 'Fo',
        # 'hundredgigabitethernet': 'Hu',
        # 'management': 'Ma',
    }

    def _get_short_if_name(interface: str) -> str:
        """
        returns short interface name from long interface name
        interface: is the long interface name
        :type interface: str
        """
        if not interface:
            return interface
        for interface_prefix in _alternate_if_name.keys():
            if interface.lower().startswith(interface_prefix.lower()):
                interface_short = _alternate_if_name[interface_prefix]
                return interface.lower().replace(interface_prefix.lower(), interface_short, 1)
        return interface

    # try to find the item for an interface
    def _match_entry_with_item(_entry: Mapping[str, str], services: Sequence[str]) -> str | None:
        values = [
            _entry.get('name'.strip()),
            _entry.get('description'.strip()),
            _entry.get('alias').strip()
        ]
        for value in values:
            if value in services:
                return value

        index = str(_entry.get('index'))

        # try alias+index
        alias_index = str(_entry.get('alias')).strip() + ' ' + index
        if alias_index in services:
            LOGGER.info(f'match found by alias-index|{_entry}|{alias_index}|')
            return alias_index

        # try descrption+index
        description_index = str(_entry.get('description')).strip() + ' ' + index
        if description_index in services:
            LOGGER.info(f'match found by alias-index|{_entry}|{description_index}|')
            return description_index

        # for index try with padding
        pad_services: List[str] = [x for x in services if x.isdigit()]
        if pad_services:
            max_pad = len(max(pad_services, key=len)) + 1
            min_pad = len(min(pad_services, key=len))
            for i in range(min_pad, max_pad):
                index_padded = f'{index:0>{i}}'
                if index_padded in pad_services:
                    return index_padded
                # still not found try values + index
                for value in values:
                    if f'{value} {index_padded}' in services:
                        return f'{value} {index_padded}'

        LOGGER.warning(f'no match found |{_entry}|{services}|')
        return None

    # empty host/neighbour should never happen here
    if not host:
        LOGGER.warning(f'no host name |{host}|')
        return None

    # get dict of interfaces with the item as key
    interface_data: Mapping[str, Mapping[str, object]] = host_cache.get_data(
        host=host, item=CacheItems.interfaces, path=CACHE_INTERFACES_DATA
    )
    if not interface_data:
        LOGGER.warning(f'got no interface data for: {host}')
        return None

    # try to find the interface in the host interface inventory list
    inventory = host_cache.get_data(
        host=host, item=CacheItems.inventory, path=PATH_INTERFACES
    )
    if not inventory:
        LOGGER.warning(f'no interface inventory for host: {host}')
        return None

    interface_items: Sequence[str] = list(interface_data.keys())

    # the easy case
    if interface in interface_items:
        return interface

    for _entry in inventory:
        if interface in [
            _entry.get('name'),
            _entry.get('description'),
            _entry.get('alias'),
            str(_entry.get('index')),
            _entry.get('phys_address'),
        ]:
            return _match_entry_with_item(_entry, interface_items)
        elif f'1:{interface}' == _entry.get('name'):  # Extreme non stack:
            return _match_entry_with_item(_entry, interface_items)
        elif _entry.get('name') is not None and _get_short_if_name(
                _entry.get('name')) == interface.lower():  # Cisco NXOS
            return _match_entry_with_item(_entry, interface_items)

    LOGGER.warning(msg=f'Device: {host}: service for interface |{interface}| not found')


def add_tooltip_html(
        metadata: Dict,
        type_: str,
        left: str,
        left_value: str,
        right: str,
        right_value: str,
) -> Dict:
    css_class = 'class=mismatch'
    if metadata.get('tooltip') is None:
        metadata['tooltip'] = {}
    if metadata['tooltip'].get('html') is None:
        css_style = (
            '<style>'
                'div.mismatch {'
                    'background-color: rgba(70,70,70,.1);'
                    'border-radius: 5px;'
                    'padding: 5px;'
                '}'
                'p.mismatch {'
                    'text-align: center;'
                '}'
                'table.mismatch {'
                    'text-align: left;'
                    'border-radius: 5px;'
                '}'
                'tr.mismatch {'
                '}'
                'tr.mismatch:nth-child(even) {'
                    'background-color: rgba(100,100,100,.3);'

                '}'
                'tr.mismatch:nth-child(odd) {'
                    'background-color: rgba(100,100,100,.2);'

                '}'
                'th.mismatch {'
                    'background-color: rgba(120,120,120,.3);'
                    'padding: 5px;'   # inside the element
                    # 'margin: 5px;'  # outside of the lement
                '}'
                'td.mismatch {'
                    'padding: 5px;'
                '}'
            '</style>'
        )
        header = (
            '<thead>'
                '<tr>'
                    f'<th {css_class}>Node</th>'
                    f'<th {css_class}>{left}</th>'
                    f'<th {css_class}>{right}</th>'
                '</tr>'
            '</thead>'
        )
        metadata['tooltip']['html'] = (
            f'{css_style}'
            f'<div {css_class}>'
                f'<p {css_class}>WARNING: Mismatch found!</p>'
                f'<table {css_class}>'
                    f'{header}'
                    '<tbody>'
        )

    metadata['tooltip']['html'] += (
        f'<tr {css_class}>'
            f'<td {css_class}>{type_}</td>'
            f'<td {css_class}>{left_value}</td>'
            f'<td {css_class}>{right_value}</td>'
        '</tr>'
    )

    return metadata


def close_tooltip_html(metadata: Dict) -> Dict:
    metadata['tooltip']['html'] += '</table></div>'
    return metadata


def get_network_summary(ipv4_address: str, summarize: Sequence[IPv4Network]) -> str | None:
    for network in summarize:
        if IPv4Network(ipv4_address).subnet_of(network):
            return network.exploded
    return None


def is_ignore_ipv4(ip_address: str, ignore_ips: Sequence[IPv4Network]) -> bool:
    for ip in ignore_ips:
        if IPv4Network(ip_address).subnet_of(ip):
            LOGGER.info(f'IP address {ip_address} is in ignore list -> ({ip})')
            return True
    return False


def is_ignore_wildcard(ip_address: str, ignore_wildcard: Sequence[Wildcard]) -> bool:
    int_ip_address = int(IPv4Address(ip_address))
    for wildcard in ignore_wildcard:
        if int_ip_address & wildcard.int_wildcard == wildcard.bit_pattern:
            LOGGER.info(
                f'IP address {ip_address} matches ignore wildcard '
                f'list ({wildcard.ip_address}/{wildcard.wildcard})'
            )
            return True
    return False


def get_list_of_devices(data: Mapping) -> List[str]:
    devices: List[str] = []
    for connection in data.values():
        devices.append(connection[0])
    return list(set(devices))
