#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# License: GNU General Public License v2

# Author: thl-cmk[at]outlook[dot]com
# URL   : https://thl-cmk.hopto.org
# Date  : 2023-10-08
# File  : nvdct_data.py

# 2023-10-10: initial release
# 2023-10-16: added options  --keep-max  and --min-age
# 2023-10-17: removed -p, -c, -d (use the long options instead: --path-in-inventory, --data-source,
#             --inventory-columns)
#             changed option --keep-max to --keep
#             added option --check-user-data-only
#             added SEED_DEVICES to user data, option -s/--sed-devices is now optional
#             refactoring
# 2023-10-18: fixed --make-default not working
#             added protected topologies
#             changed get inventory data from file access to request via live status query
#             get interfaces from autochecks changed to live status query
# 2023-10-19: cleanup removed "all" unused functions
#             added option -u, --user-data-file to provide a user specific data file
# 2023-10-20: changed option -m/--make-default to -d/--default (-m is needed for -m/--merge
#             added -m/--merge option
# 2023-10-27: improved matching of port name from CDP/LLDP data to interface (item) name of service
#             improved internal data handling (HostCache)
# 2023-10-28: reworked handling of layers (see option -l/--layers)
#             added option -l/-layers {CDP,CUSTOM,LLDP,STATIC} [{CDP,CUSTOM,LLDP,STATIC} ...]
#             added section CUSTOM_LAYERS to user data file
#             removed option -m/--merge, included in option -l/--layers
#             removed option --lldp, included in option -l/--layers
#             removed option --data-source, now handled in CUSTOM_LAYERS
#             removed option --inventory-columns, now handled in CUSTOM_LAYERS
#             removed option --path-in-inventory, now handled in CUSTOM_LAYERS
#             removed lower limits for --keep and --min-age
# 2023-1ß-29: fixed missing path from custom layer in host prefetch
# 2023-11-16: added option -b/--backend [LIVESTATUS, FILESYSTEM],
#             LIVESTATUS is the default for performance reasons -> for now only local site
#             FILESYSTEM fetches the data directly form the inventory files -> use in distributed
#                        environments
# 2023-11-28: implemented MULTISITE as backend, via livestatus -> use in distributed environments,
#             may become the default
# 2023-11-30: added RESTAPI backend --> this is kind of slow (20 seconds compared to 0.5
#             for MULTISITE)
# 2023-12-06: do not add devices from static connections to seed devices --> be more transparent
#             added prefix option -> needs testing (THX to andreas doehler for the suggestion)
# 2023-12-21: adjusted for streamlined CDP/LLDP inventory format,
#             needs at least CDP 0.7.1-20231221 and LLDP 0.9.2-2023122
#             changed default backend from LIVESTATUS to MULTISITE
#             added logger (option --log-level)
#             removed option --debug
# 2023-12-22: added options --log-file and --log-to-stdout
# 2024-01-02: added new output format (option --new-format)
#             fixed LOGGER.warn --> LOGGER.warning
#             improved detection of missing hosts/interfaces
# 2024-01-03: fixed query in MULTISITE.exist
#             added option --quiet
#             renamed from create_topology_data to nvdct
#             added layer3 IPv4 topology (experimental)
# 2024-01-04: added SETTINGS section to nvdct.toml file -> defaults -> overridden by
#             toml -> overridden by cli
#             only seed devices are additive (toml + cli)
#             added L3v4 toml options: L3V4_IGNORE_HOSTS, L3V4_SUMMARIZE, L3V3_REPLACE
#             moved from ~/local/bin/topology_data to ~/local/bin/nvdct
# 2024-01-28: fixed -b MULTISITE if not in distributed environment (falling back to local)
#             fixed crash in create_l3v4_topology if no host was found
# 2024-01-29: added new output path for cmk 2.3
#             changed don't output cmk 2.2 format if --new-format set
# 2024-02-25: added images for L3 Topo (IP-Address and IP-Network)
#             icon source includes: https://www.flaticon.com/free-icons
#             added option to drop neighbours by regex
#             fixed no inventory in L3v4
# 2024-03-14: added metadata for connection (thickness, color)
#             thickness can be configured via toml (MAP_SPEED_TO_THICKNESS)
#             color 'red' is used for speed mismatch
#             made node emblem configurable (for ip-address and ip-network) via toml (EMBLEMS)
# 2024-03-16: moved args, utils, classes to lib
#             split classes in to backends, settings and classes
# 2024-03-17: changed speed mismatch from native speed to operational speed
#             added duplex mismatch ('yellow')
#             added native vlan mismatch ('orange')
# 2024-03-25: added --pre-fetch, this will speed up the RESTAPI backend performance
#             min inv_cdp_cahe version: 0.7.1-20240320 -> host label nvdct/has_cdp_neighbours
#             min inv_lldp_cache version: 0.9.3-20240320 -> host label nvdct/has_lldp_neighbours
# 2024-03-27: added option --api-port, defaults to 80
# 2024-03-28: changed restapi get_interface_data to use one call to fetch all data
# 2024-03-29: added L3V4_IGNORE_IP and L3V4_IGNORE_WILDCARD to toml file
# 2024-03-31: fixed: L3V3_REPLACE -> L3V4_REPLACE
#             made input from toml more bullet proof
# 2024-04-03: change HostCacheMultiSite to be child of HostCacheLiveStatus
# 2024-04-06: added node icon warning
# 2024-04-07: added css for connection
# 2024-04-13: added html tooltip for connection issues
# 2024-04-15: added emblems for l3v4_summary and l3v4_replace
# 2024-04-16: fixed html tooltip for speed mismatch (Speed: 100 MBit/s (expected: 1 GBit/s)WARN)
#             added option --skip-l3-if
# 2024-04-17: automatically detect local CMK version -> no need for --new-format option any more
#             automatically detect local CONFIG_APACHE_TCP_PORT -> no need for --api-port option any more  # noqa: E501
# 2024-04-18: added option --skip-l3-ip
#             added tooltip for --skip-l3-if and --skip-l3-ip
# 2024-04-30: automatically create @default if it doesn't exist
# 2024-05-07: fixed handling of ignore wildcard in L3v4 topology
# 2024-05-17: added support for Extreme interfaces in LLDP without stack id (1 -> 1:1)
# 2024-05-18: added support for NXOS interfaces in LLDP in short from (Eth1/1 -> Ethernet1/1)
#             added IP-Address/IP-Network as quickinfo
# 2024-05-18: fixed crash non empty neighbour port (ThX to andreas doehler)
# 2024-06-05: fixed interface index padding
# 2024-06-06: added interfcae detection alias+index
#             added interface detection description + index
# 2024-06-09: moved topologiy helpers to lib/topologies.py
#             moved (default) config file(s) to ./conf/
# 2024-06-14: added debug code for bad IPv4 address data
# 2024-06-17: fixed bad IPv4 address data (just drop it)

# creating topology data json from inventory data
#
# This script creates the topology data file needed for the Checkmk "network_visualization" plugin
# by Andreas Boesl/schnetz. For more information see
# https://forum.checkmk.com/t/network-visualization/41680  (from 2023-10-05)
# https://exchange.checkmk.com/p/network-visualization     (from 2023-10-05)
#
# NOTE: the topology_data configuration (layout etc.) is saved under ~/var/check_mk/topology
#
# The inventory data could be created with my CDP/LLDP/IP Address/Interface nane inventory plugins:
# CDP.....: https://thl-cmk.hopto.org/gitlab/checkmk/vendor-independent/inventory/inv_cdp_cache
# LLDP....: https://thl-cmk.hopto.org/gitlab/checkmk/vendor-independent/inventory/inv_lldp_cache
# L3v4....: https://thl-cmk.hopto.org/gitlab/checkmk/vendor-independent/inventory/inv_ipv4_addresses
# IF Name.: https://thl-cmk.hopto.org/gitlab/checkmk/vendor-independent/inventory/inv_ifname
#
# USAGE:
# ~/local/lib/nvdct/nvdct.py --help
#

"""
# ##################################################################################################
# ##################################################################################################
# ##################################### data format ################################################
# ##################################################################################################
# ##################################################################################################

__objects = {
    "HostA": {
        "name": "myhost",
        "link": {"core": "myhost"},
        'metadata': {},
    },
    "Service@HostA": {
        "name": "SomeService",
        "link": {"core": ["myhost", "interface3"]},  # service@host in CMK
        # "link": {"core": "myhost"},                # host object in CMK
        # "link": {},                                # object not in CMK core
        'metadata': {
            'images': {
                'icon': 'icon_warning',              # additional image shown at the node
                'emblem': 'ip-network_80',           # image to show for an object
            },
            'tooltip': {
                'html': "Test message<div style='color: blue'>Some color <b>test</b></div>",
                'quickinfo': [
                                {'name': "Duplex", 'value': 'full duplex'},
                                {'name': 'Speed', 'value': '123132234'},
                                {'name': 'Native vlan', 'value': '123'}
                            ]
            }
        },
    },
}

__connections = [
    [["HostA", "ServiceOfHostA"], {}],
    [["ServiceOfHostA", "ServiceOfHostB"], {
        'line_config': {
            'thickness': 3,
            'color': '332211',
            'css_styles': {
                'stroke': 'green',
                'stroke-dasharray': '30 30',
                # 'stroke-dasharray': 'unset',  # to remove a CSS attribute
                'stroke-width': '4',
            }
        },
        'tooltip': {
            'html': "Test message mismatch<div style='color: blue'>Some color <b>test</b></div>",
            'quickinfo': [
                            {'name': "Duplex", 'value': 'full duplex'},
                            {'name': 'Native speed', 'value': '123132234'},
                            {'name': 'Native vlan', 'value': '123'}
                        ]
        }
    }],
    [["ServiceOfHostB", "HostB"], {}],
]

__data = {
    'version': 1,
    'name': 'LLDP',  # label for this layer
    'objects': __objects,
    'connections': __connections,
}

# ##################################################################################################
# ##################################################################################################
# ##################################### data format ################################################
# ##################################################################################################
# ##################################################################################################
"""

import sys
from _collections_abc import Mapping, Sequence
from ipaddress import IPv4Network
from logging import DEBUG
from re import compile as re_compile
from time import strftime, time_ns
from typing import Dict, List

from lib.args import parse_arguments
from lib.backends import (
    CacheItems,
    HostCache,
    HostCacheFileSystem,
    HostCacheLiveStatus,
    HostCacheMultiSite,
    HostCacheRestApi,
)
from lib.topologies import (
    NvConnections,
    NvObjects,
    get_list_of_devices,
    get_network_summary,
    get_service_by_interface,
    is_ignore_ipv4,
    is_ignore_wildcard,
)
from lib.utils import (
    configure_logger,
    ExitCodes,
    HOME_URL,
    HOST_LABEL_L3V4,
    InventoryColumns,
    Ipv4Info,
    Layer,
    LAYERS,
    LOGGER,
    merge_topologies,
    NVDCT_VERSION,
    PATH_L3v4,
    remove_old_data,
    save_data_to_file,
    save_topology,
    StdoutQuiet,
)
from lib.settings import (
    Emblems,
    Settings,
    StaticConnection,
    Thickness,
    Wildcard,
)


DROP_HOSTS_REGEX: List[str] = []
DROP_HOSTS: List[str] = []
EMBLEMS: Emblems
HOST_CACHE: HostCache
HOST_MAP: Dict[str, str] = {}
MAP_SPEED_TO_THICKNESS: List[Thickness] = []
NV_CONNECTIONS = NvConnections()
NV_OBJECTS = NvObjects()
SETTINGS: Settings


def create_l2_device_from_inv(
        host: str,
        inv_data: Sequence[Mapping[str, str]],
        inv_columns: InventoryColumns,
        label: str,
) -> Dict[str, object] | None:
    data: Dict = {'connections': {}, "interfaces": []}

    for topo_neighbour in inv_data:
        # check if required data are not empty
        if not (neighbour := topo_neighbour.get(inv_columns.neighbour)):
            continue
        if not (raw_local_port := topo_neighbour.get(inv_columns.local_port)):
            continue
        if not (raw_neighbour_port := topo_neighbour.get(inv_columns.neighbour_port)):
            continue

        # drop neighbour before domain split
        if neighbour in DROP_HOSTS:
            LOGGER.info(msg=f'drop neighbour: {neighbour}')
            continue

        drop_regex = False
        for re_str in DROP_HOSTS_REGEX:
            re_host = re_compile(re_str)
            if re_host.match(neighbour):
                LOGGER.info(msg=f'regex drop neighbour: {neighbour}, regex str: {re_str}')
                drop_regex = True
                break
        if drop_regex:
            continue

        if not SETTINGS.keep_domain:
            neighbour = neighbour.split('.')[0]

        # drop neighbour after domain split
        if neighbour in DROP_HOSTS:
            LOGGER.info(msg=f'drop neighbour: {neighbour}')
            continue

        if SETTINGS.uppercase:
            neighbour = neighbour.upper()
        if SETTINGS.lowercase:
            neighbour = neighbour.lower()
        if SETTINGS.prefix:
            neighbour = f'{SETTINGS.prefix}{neighbour}'
        # rewrite neighbour if inventory neighbour and checkmk host don't match
        if neighbour in HOST_MAP.keys():
            neighbour = HOST_MAP[neighbour]

        # getting/checking interfaces
        local_port = get_service_by_interface(host, raw_local_port, HOST_CACHE)
        if not local_port:
            local_port = raw_local_port
            LOGGER.warning(msg=f'service not found: host: {host}, raw_local_port: {raw_local_port}')
        elif local_port != raw_local_port:
            # local_port = raw_local_port  # don't reset local_port
            LOGGER.info(
                msg=f'host: {host}, raw_local_port: {raw_local_port} -> local_port: {local_port}'
            )

        neighbour_port = get_service_by_interface(neighbour, raw_neighbour_port, HOST_CACHE)
        if not neighbour_port:
            neighbour_port = raw_neighbour_port
            LOGGER.warning(
                msg=f'service not found: neighbour: {neighbour}, '
                    f'raw_neighbour_port: {raw_neighbour_port}'
            )
        elif neighbour_port != raw_neighbour_port:
            # neighbour_port = raw_neighbour_port  # don't reset neighbour_port
            LOGGER.info(
                msg=f'neighbour: {neighbour}, raw_neighbour_port {raw_neighbour_port} '
                    f'-> neighbour_port {neighbour_port}'
            )

        if neighbour and local_port and neighbour_port:
            data['connections'].update({local_port: [neighbour, neighbour_port, label]})
            if local_port not in data['interfaces']:
                data['interfaces'].append(local_port)

        metadata = {
            'duplex': topo_neighbour.get('duplex'),
            'native_vlan': topo_neighbour.get('native_vlan'),
        }

        # add to new object list
        NV_OBJECTS.add_host_object(host=host, host_cache=HOST_CACHE)
        NV_OBJECTS.add_host_object(host=neighbour, host_cache=HOST_CACHE)
        NV_OBJECTS.add_service_object(
            host=host,
            service=local_port,
            host_cache=HOST_CACHE,
            metadata=metadata,
            name=raw_local_port,
            item=local_port
        )
        NV_OBJECTS.add_service_object(
            host=neighbour,
            service=neighbour_port,
            host_cache=HOST_CACHE,
            name=raw_neighbour_port,
            item=neighbour_port
        )
        NV_CONNECTIONS.add_connection(
            left=host,
            right=f'{local_port}@{host}',
        )
        NV_CONNECTIONS.add_connection(
            left=neighbour,
            right=f'{neighbour_port}@{neighbour}',
        )
        NV_CONNECTIONS.add_connection(
            left=f'{local_port}@{host}',
            right=f'{neighbour_port}@{neighbour}',
        )

    return {host: data}


def create_static_connections(connections: Sequence[StaticConnection]) -> Dict:
    data: Dict = {}
    for connection in connections:
        LOGGER.info(msg=f'connection: {connection}')
        if connection.host not in data:
            data[connection.host] = {'connections': {}, 'interfaces': []}

        if connection.neighbour not in data:
            data[connection.neighbour] = {'connections': {}, 'interfaces': []}

        # add connection from host to neighbour
        data[connection.host]['connections'].update({connection.local_port: [
            connection.neighbour, connection.neighbour_port, connection.label
        ]})
        data[connection.host]['interfaces'].append(connection.local_port)

        # add connection from neighbour to host
        data[connection.neighbour]['connections'].update({
            connection.neighbour_port: [connection.host, connection.local_port, connection.label]
        })
        data[connection.neighbour]['interfaces'].append(connection.neighbour_port)

        # new format
        NV_OBJECTS.add_host_object(
            host=connection.host,
            host_cache=HOST_CACHE,
            emblem=EMBLEMS.host_node
        )
        NV_OBJECTS.add_host_object(
            host=connection.neighbour,
            host_cache=HOST_CACHE,
            emblem=EMBLEMS.host_node
        )
        NV_OBJECTS.add_service_object(
            host=connection.host,
            host_cache=HOST_CACHE,
            emblem=EMBLEMS.service_node,
            service=connection.local_port
        )
        NV_OBJECTS.add_service_object(
            host=connection.neighbour,
            host_cache=HOST_CACHE,
            emblem=EMBLEMS.service_node,
            service=connection.neighbour_port
        )
        NV_CONNECTIONS.add_connection(
            left=connection.host,
            right=f'{connection.local_port}@{connection.host}',
        )
        NV_CONNECTIONS.add_connection(
            left=connection.neighbour,
            right=f'{connection.neighbour_port}@{connection.neighbour}',
        )
        NV_CONNECTIONS.add_connection(
            left=f'{connection.local_port}@{connection.host}',
            right=f'{connection.neighbour_port}@{connection.neighbour}',
        )

    LOGGER.debug(msg=data)
    LOGGER.info(msg=f'Devices added: {len(data)}, source STATIC')
    # if not SETTINGS.quiet:
    print(f'Devices added.: {len(data)}, source STATIC')

    return data


def create_l2_topology(
        seed_devices: Sequence[str],
        path_in_inventory: str,
        inv_columns: InventoryColumns,
        label: str,
) -> Dict:
    devices_to_go = list(set(seed_devices))  # remove duplicates
    topology_data: Dict = {}
    devices_done = []

    while devices_to_go:
        device = devices_to_go[0]

        if device in HOST_MAP.keys():
            try:
                devices_to_go.remove(device)
            except ValueError:
                pass
            device = HOST_MAP[device]
            if device in devices_done:
                continue

        topo_data = HOST_CACHE.get_data(
            host=device, item=CacheItems.inventory, path=path_in_inventory
        )
        if topo_data:
            topology_data.update(
                create_l2_device_from_inv(
                    host=device,
                    inv_data=topo_data,
                    inv_columns=inv_columns,
                    label=label,
                ))
            devices_list = get_list_of_devices(topology_data[device]['connections'])
            for _entry in devices_list:
                if _entry not in devices_done:
                    devices_to_go.append(_entry)

        devices_to_go = list(set(devices_to_go))
        devices_done.append(device)
        devices_to_go.remove(device)
        LOGGER.info(msg=f'Device done: {device}, source: {label}')

    LOGGER.info(f'Devices added: {len(devices_done)}, source {label}')
    # if not SETTINGS.quiet:
    print(f'Devices added.: {len(devices_done)}, source {label}')

    return topology_data


def create_l3v4_topology(
        ignore_hosts: Sequence[str],
        ignore_ips: Sequence[IPv4Network],
        ignore_wildcard: Sequence[Wildcard],
        summarize: Sequence[IPv4Network],
        replace: Mapping[str, str],
        skip_if: bool,
        skip_ip: bool,
) -> None:
    host_list: Sequence[str] = HOST_CACHE.get_hosts_by_label(HOST_LABEL_L3V4)
    LOGGER.debug(f'host list: {host_list}')
    if not host_list:
        LOGGER.warning(
            msg='No routing capable host found. Check if "inv_ipv4_addresses.mkp" '
                'added/enabled and inventory has run.'
        )
        return

    LOGGER.debug(f'L3v4 ignore hosts: {ignore_hosts}')
    for raw_host in host_list:
        host = raw_host
        if host in ignore_hosts:
            LOGGER.info(f'L3v4 host {host} ignored')
            continue
        if not (ipv4_addresses := HOST_CACHE.get_data(
            host=host, item=CacheItems.inventory, path=PATH_L3v4)
        ):
            LOGGER.warning(f'No IPv4 address inventory found for host: {host}')
            continue

        NV_OBJECTS.add_host_object(host=host, host_cache=HOST_CACHE)
        for _entry in ipv4_addresses:
            emblem = EMBLEMS.ip_network
            try:
                ipv4_info = Ipv4Info(**_entry)
            except TypeError as e:
                LOGGER.warning(f'Drop IPv4 address data for host: {host}, data: {_entry}')
                continue

            if ipv4_info.address.startswith('127.'):  # drop loopback addresses
                LOGGER.info(f'host: {host} dropped loopback address: {ipv4_info.address}')
                continue

            if ipv4_info.cidr == 32:  # drop host addresses
                LOGGER.info(f'host: {host} dropped host address: {ipv4_info.address}/{ipv4_info.cidr}')
                continue

            if ipv4_info.type.lower() != 'ipv4':  # drop if not ipv4
                LOGGER.warning(f'host: {host} dropped non ipv4 address: {ipv4_info.address}, type: {ipv4_info.type}')
                continue

            if is_ignore_ipv4(ipv4_info.address, ignore_ips):
                LOGGER.info(f'host: {host} dropped ignore address: {ipv4_info.address}')
                continue

            if is_ignore_wildcard(ipv4_info.address, ignore_wildcard):
                LOGGER.info(f'host: {host} dropped wildcard address: {ipv4_info.address}')
                continue

            if network := get_network_summary(
                ipv4_address=ipv4_info.address,
                summarize=summarize,
            ):
                emblem = EMBLEMS.l3v4_summarize
                LOGGER.info(
                    f'Network summarized: {ipv4_info.network}/{ipv4_info.cidr} -> {network}'
                )
            else:
                network = f'{ipv4_info.network}/{ipv4_info.cidr}'

            if network in replace.keys():
                LOGGER.info(f'Replaced network {network} with {replace[network]}')
                network = replace[network]
                emblem = EMBLEMS.l3v4_replace

            NV_OBJECTS.add_ipv4_network_object(network=network, emblem=emblem)

            if skip_if is True and skip_ip is True:
                NV_CONNECTIONS.add_connection(left=host, right=network)
            elif skip_if is True and skip_ip is False:
                NV_OBJECTS.add_ipv4_address_object(
                    host=host,
                    interface=None,
                    ipv4_address=ipv4_info.address,
                    emblem=EMBLEMS.ip_address,
                )
                NV_OBJECTS.add_tooltip_quickinfo(
                    '{ipv4_info.address}@{host}', 'Interface', ipv4_info.device
                )
                NV_CONNECTIONS.add_connection(left=f'{host}', right=f'{ipv4_info.address}@{host}')
                NV_CONNECTIONS.add_connection(left=network, right=f'{ipv4_info.address}@{host}')
            elif skip_if is False and skip_ip is True:
                NV_OBJECTS.add_service_object(
                    host=host, service=ipv4_info.device, host_cache=HOST_CACHE
                )
                NV_OBJECTS.add_tooltip_quickinfo(
                    f'{ipv4_info.device}@{host}', 'IP-address', ipv4_info.address
                )
                NV_CONNECTIONS.add_connection(left=f'{host}', right=f'{ipv4_info.device}@{host}')
                NV_CONNECTIONS.add_connection(left=network, right=f'{ipv4_info.device}@{host}')
            else:
                NV_OBJECTS.add_ipv4_address_object(
                    host=host,
                    interface=ipv4_info.device,
                    ipv4_address=ipv4_info.address,
                    emblem=EMBLEMS.ip_address,
                )
                NV_OBJECTS.add_service_object(
                    host=host, service=ipv4_info.device, host_cache=HOST_CACHE,
                )
                NV_CONNECTIONS.add_connection(
                    left=host, right=f'{ipv4_info.device}@{host}')
                NV_CONNECTIONS.add_connection(
                    left=f'{ipv4_info.device}@{host}',
                    right=f'{ipv4_info.address}@{ipv4_info.device}@{host}',
                )
                NV_CONNECTIONS.add_connection(
                    left=network, right=f'{ipv4_info.address}@{ipv4_info.device}@{host}',
                )

    LOGGER.info(msg=f'Devices added: {len(host_list)}, source L3v4')
    # if not SETTINGS.quiet:
    print(f'Devices added.: {len(host_list)}, source L3v4')


if __name__ == '__main__':
    start_time = time_ns()
    SETTINGS = Settings(vars(parse_arguments()))
    sys.stdout = StdoutQuiet(quiet=SETTINGS.quiet)

    configure_logger(
        log_file=SETTINGS.log_file,
        log_to_console=SETTINGS.log_to_stdtout,
        log_level=SETTINGS.loglevel,
    )
    LOGGER.info(msg='Data creation started')

    # if not SETTINGS.quiet:
    print()
    print(
        f'Network Visualisation Data Creation Tool (NVDCT)\n'
        f'by thl-cmk[at]outlook[dot]com, version {NVDCT_VERSION}\n'
        f'see {HOME_URL}'
    )
    print()
    print(f'Start time....: {strftime(SETTINGS.time_format)}')

    _backends = {
        'LIVESTATUS': HostCacheLiveStatus,
        'FILESYSTEM': HostCacheFileSystem,
        'MULTISITE': HostCacheMultiSite,
        'RESTAPI': HostCacheRestApi,
    }
    host_cache_backend = _backends.get(SETTINGS.backend, None)
    if not host_cache_backend:
        LOGGER.error(msg=f'Backend {SETTINGS.backend} not (yet) implemented')
        sys.exit(ExitCodes.BACKEND_NOT_IMPLEMENTED.value)
    elif SETTINGS.backend == 'RESTAPI':
        HOST_CACHE = host_cache_backend(
            pre_fetch=SETTINGS.pre_fetch,
            api_port=SETTINGS.api_port
        )
    else:
        HOST_CACHE = host_cache_backend(
            pre_fetch=SETTINGS.pre_fetch,
        )

    HOST_MAP = SETTINGS.host_map
    DROP_HOSTS = SETTINGS.drop_hosts
    DROP_HOSTS_REGEX = SETTINGS.drop_hosts_regex
    MAP_SPEED_TO_THICKNESS = SETTINGS.map_speed_to_thickness
    EMBLEMS = SETTINGS.emblems

    jobs: List[Layer] = []
    final_topology: Dict = {}
    pre_fetch_layers: List[str] = []
    pre_fetch_host_list: List[str] = []

    for layer in SETTINGS.layers:
        if layer == 'STATIC':
            jobs.append(layer)
        if layer == 'L3v4':
            jobs.append(layer)
            HOST_CACHE.add_inventory_prefetch_path(path=LAYERS[layer].path)
            pre_fetch_layers.append(LAYERS[layer].host_label)

        elif layer in LAYERS:
            jobs.append(LAYERS[layer])
            HOST_CACHE.add_inventory_prefetch_path(path=LAYERS[layer].path)
            pre_fetch_layers.append(LAYERS[layer].host_label)

        elif layer == 'CUSTOM':
            for entry in SETTINGS.custom_layers:
                jobs.append(entry)
                HOST_CACHE.add_inventory_prefetch_path(entry.path)

    if SETTINGS.pre_fetch:
        LOGGER.info('Pre fill cache...')
        for host_label in pre_fetch_layers:
            if _host_list := HOST_CACHE.get_hosts_by_label(host_label):
                pre_fetch_host_list = list(set(pre_fetch_host_list + _host_list))
        LOGGER.info(f'Fetching data for {len(pre_fetch_host_list)} hosts start')
        print(f'Prefetch start: {strftime(SETTINGS.time_format)}')
        print(f'Prefetch hosts: {len(pre_fetch_host_list)} of {len(HOST_CACHE.cache.keys())}')
        HOST_CACHE.pre_fetch_cache(pre_fetch_host_list)
        LOGGER.info(f'Fetching data for {len(pre_fetch_host_list)} hosts end')
        print(f'Prefetch end..: {strftime(SETTINGS.time_format)}')

    for job in jobs:
        match job:
            case 'STATIC':
                topology: Dict | None = create_static_connections(
                    connections=SETTINGS.static_connections
                )
                label = 'static'
            case 'L3v4':
                topology = None
                label = 'l3v4'
                if SETTINGS.new_format:
                    create_l3v4_topology(
                        ignore_hosts=SETTINGS.l3v4_ignore_hosts,
                        ignore_ips=SETTINGS.l3v4_ignore_ips,
                        ignore_wildcard=SETTINGS.l3v4_ignore_wildcard,
                        summarize=SETTINGS.l3v4_summarize,
                        replace=SETTINGS.l3v4_replace,
                        skip_if=SETTINGS.skip_l3_if,
                        skip_ip=SETTINGS.skip_l3_ip
                    )
            case _:
                label = job.label.lower()
                columns = job.columns.split(',')
                topology = create_l2_topology(
                    seed_devices=SETTINGS.seed_devices,
                    path_in_inventory=job.path,
                    inv_columns=InventoryColumns(
                        neighbour=columns[0],
                        local_port=columns[1],
                        neighbour_port=columns[2]
                    ),
                    label=label,
                )

        if SETTINGS.new_format:
            LOGGER.info(f'Save topology {label} in new format')
            NV_CONNECTIONS.add_meta_data_to_connections(
                nv_objects=NV_OBJECTS,
                speed_map=MAP_SPEED_TO_THICKNESS,
            )
            if not SETTINGS.loglevel == DEBUG:
                connections = NV_CONNECTIONS.nv_connections
            else:
                connections = sorted(NV_CONNECTIONS.nv_connections)
            _data = {
                'version': 1,
                'name': label,
                'objects': NV_OBJECTS.nv_objects if not SETTINGS.loglevel == DEBUG else dict(
                    sorted(NV_OBJECTS.nv_objects.items())
                ),
                'connections': connections
            }
            save_data_to_file(
                data=_data,
                path=(
                    f'{SETTINGS.omd_root}/{SETTINGS.topology_save_path}/'
                    f'{SETTINGS.output_directory}'
                ),
                file=f'data_{label}.json',
                make_default=SETTINGS.default,
            )

        NV_OBJECTS = NvObjects()
        NV_CONNECTIONS = NvConnections()

        if topology:
            final_topology = merge_topologies(topo_pri=topology, topo_sec=final_topology)

    if final_topology and not SETTINGS.new_format:
        LOGGER.info('Save topology in CMK 2.2.x format')

        save_topology(
            data=final_topology,
            base_directory=f'{SETTINGS.omd_root}/{SETTINGS.topology_save_path}',
            output_directory=SETTINGS.output_directory,
            topology_file_name=SETTINGS.topology_file_name,
            dont_compare=SETTINGS.dont_compare,
            make_default=SETTINGS.default,
        )

    if SETTINGS.keep:
        remove_old_data(
            keep=SETTINGS.keep,
            min_age=SETTINGS.min_age,
            raw_path=f'{SETTINGS.omd_root}/{SETTINGS.topology_save_path}',
            protected=SETTINGS.protected_topologies,
        )

    # if not SETTINGS.quiet:
    print(f'Time taken....: {(time_ns() - start_time) / 1e9}/s')
    print(f'End time......: {strftime(SETTINGS.time_format)}')
    print()

    LOGGER.info('Data creation finished')
